/*
 *  Licensed to the Apache Software Foundation (ASF) under one
 *  or more contributor license agreements.  See the NOTICE file
 *  distributed with this work for additional information
 *  regarding copyright ownership.  The ASF licenses this file
 *  to you under the Apache License, Version 2.0 (the
 *  "License"); you may not use this file except in compliance
 *  with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing,
 *  software distributed under the License is distributed on an
 *   * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 *  KIND, either express or implied.  See the License for the
 *  specific language governing permissions and limitations
 *  under the License.
 */

package org.wso2.registry;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.wso2.registry.jdbc.dao.VersionedResourceDAO;

import javax.sql.DataSource;
import java.sql.Timestamp;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.*;
import java.io.*;

/**
 * Represents any file or collection stored in the registry. It encapsulates both the content
 * of the entity and its metadata. In addition to files and collections, instances of Resource
 * are used to represent results of runtime queries. In such secenarios, comments, tags, ratings as
 * well as collections of comments, etc. are also represented by Resource objects.
 *
 * Each resource instance contains a unique path within a Registry instance. Registry.get(...)
 * method invocation using this path gives an instance of that resource. This path can be
 * combined with the base URL of the registry server to generate a URI for the resource.
 */
public class ResourceImpl implements Resource {

    private static final Log log = LogFactory.getLog(ResourceImpl.class);

    /**
     * Unique identifier of the resource within the registry. Note that this is not an UUID.
     */
    private long id;

    /**
     * Version of this instance of the resource. Versioned attributes (e.g. content,
     * last modified date) can be accessed by using the combination of id and versionNumber.
     */
    private long versionNumber;

    /**
     * Username of the user who added the resource to the registry.
     */
    private String authorUserName;

    /**
     * Time at which the resource is first added to the registry.
     */
    private Timestamp createdTime;

    /**
     * Username of the user who modified the resource most recently.
     */
    private String lastUpdaterUserName;

    /**
     * Time at which the resource modified most recently.
     */
    private Timestamp lastModified;

    /**
     * Description about the resource. This may contain any text including HTML fragments.
     */
    private String description;

    /**
     * Unique path of the resource within the registry. This is generated by appending all
     * ascendant paths to the resource name separated by "/". For example assume that the resource
     * name is "users.xml". config.xml if inside the collection named "config". config collection
     * is inside the root level collection named "servers". Then the path of the resource is
     * /servers/config/users.xml.
     */
    private String path;

    /**
     * Media type of the resource. Each resource can have a media type associated with it. This can
     * be either a standart MIME media type or a custom media type defined by the users of the
     * registry. Media type is used to activate media type handlers defined in the registry.
     * Thus, by defining a media type for a resource and by registering a media type handler to
     * handle that media type, it is possible to apply special processing for resources.
     */
    private String mediaType;

    /**
     * Path of the parent collection of the resource. If the resource path is
     * /servers/config/users.xml, parent path is /servers/config.
     */
    private String parentPath;

    /**
     * Used to detect whether the resource content is modified after it is retrieved from the
     * Registry. If this is set to true at the time of adding the resource back to the Registry, new
     * version will be created.
     */
    private boolean contentModified;

    /**
     * Normal resources have the state RegistryConstants.ACTIVE_STATE (100) Deleted resources have
     * the state RegistryConstants.DELETED_STATE (101)
     */
    private int state;

    /**
     * Properties associated with the resource. A resource can contain zero or more properties,
     * where each property is a name->value pair. Both name and the value should be strings.
     */
    protected Properties properties = new Properties();

    /**
     * Paths of the resources on which this resource depends on. This feature is not implemented
     * currently.
     */
    protected List<String> dependencies = new ArrayList();

    /**
     * Paths of the resources that depends on this resource. This feature is not implemented
     * currently.
     */
    protected List<String> dependents = new ArrayList();

    /**
     * Content of the resource. Object and the type stored in this field depends on the resource
     * type. If the resource is a file with no special media type handling, this contains an array
     * of bytes (byte[]) conatining the raw bytes of the file. If the resource is a collection, this
     * contains a String[] containing the paths of child resources. If the resource is processed
     * by a media type handler, it is upto the media type handler to set a content. In that case
     * content can be anything ranging from String to custom type. Therefore, clients of the API
     * should be aware of the media type and the content for that media type.
     */
    protected Object content;

    /**
     * Specified whether the resource is a collection (directory) or a file. By default all
     * resources are considered as files. If you want to make a collection, you have to explicitly
     * set the directory field to true.
     */
    private boolean directory = false;

    /**
     * Content of the resource, represented as a input stream.
     */
    private InputStream contentStream;

    /**
     * Registry datasource. This is to be used only by the resource implemetation and users of the
     * resource are not needed to use this. Some attributes of the resource (e.g. content) is
     * fetched upon the first request to optimize the response time. Getters of such attributes use
     * this datasource to fetch the values from the database.
     */
    private DataSource dataSource;

    /**
     * Resource DAO instance to access the database directly. Note that the VersionedResourceDAO is
     * thread safe, so it is possible to keep a single instance for the resource.
     */
    private VersionedResourceDAO resourceDAO;

    public ResourceImpl() {
    }

    public void setDataSource(DataSource dataSource) {
        this.dataSource = dataSource;
        resourceDAO = new VersionedResourceDAO(dataSource);
    }

    public long getId() {
        return id;
    }

    public void setId(long id) {
        this.id = id;
    }

    public long getVersionNumber() {
        return versionNumber;
    }

    public void setVersionNumber(long versionNumber) {
        this.versionNumber = versionNumber;
    }

    public String getAuthorUserName() {
        return authorUserName;
    }

    public void setAuthorUserName(String authorUserName) {
        this.authorUserName = authorUserName;
    }

    public Date getCreatedTime() {
        return createdTime;
    }

    public void setCreatedTime(Date createdTime) {
        this.createdTime = new Timestamp(createdTime.getTime());
    }

    public Date getLastModified() {
        return lastModified;
    }

    public void setLastModified(Date lastModified) {
        this.lastModified = new Timestamp(lastModified.getTime());
    }

    public String getDescription() {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public String getPath() {
        return path;
    }

    public void setPath(String path) {
        this.path = preparePath(path);
    }

    public String getMediaType() {
        return mediaType;
    }

    public void setMediaType(String mediaType) {
        this.mediaType = mediaType;
    }

    public String getParentPath() {
        return parentPath;
    }

    public void setParentPath(String parentPath) {
        this.parentPath = parentPath;
    }

    public int getState() {
        return state;
    }

    public void setState(int state) {
        this.state = state;
    }

    public String getProperty(String key) {
        
        List propValues = (List) properties.get(key);
        if (propValues == null) {
            return null;
        } else {
            return (String) propValues.get(0);
        }
    }

    public List getPropertyValues(String key) {
        return (List) properties.get(key);
    }

    public Properties getProperties() {
        return properties;
    }

    public void setProperty(String key, String value) {

        List propValues = new ArrayList();
        propValues.add(value);
        properties.put(key, propValues);
    }

    public void addProperty(String key, String value) {

        List propValues = (List) properties.get(key);
        if (propValues != null) {
            if (!propValues.contains(value)) {
                propValues.add(value);
            }
        } else {
            propValues = new ArrayList();
            propValues.add(value);
            properties.put(key, propValues);
        }
    }

    public void setProperties(Properties properties) {
        this.properties = properties;
    }

    public InputStream getContentStream() throws RegistryException {

        if (contentStream == null) {
            if (content != null) {

                if (content instanceof byte[]) {
                    this.contentStream =
                            new BufferedInputStream(new ByteArrayInputStream((byte[]) content));

                } else if (content instanceof String) {
                    byte[] contentBytes = ((String) content).getBytes();
                    this.contentStream =
                            new BufferedInputStream(new ByteArrayInputStream(contentBytes));
                }
                
            } else if (dataSource != null) {

                Connection conn = null;

                try {
                    conn = dataSource.getConnection();
                    contentStream = resourceDAO.getResourceContentStream(id, versionNumber, conn);

                } catch (Exception e) {
                    String msg =
                            "Failed to get the input stream for the content of resource: " + path;
                    log.error(msg, e);
                    throw new RegistryException(msg, e);

                } finally {
                    if (conn != null) {
                        try {
                            conn.close();
                        } catch (SQLException e) {
                            log.error("Failed to close the database connection.", e);
                        }
                    }
                }
            }
        }

        return contentStream;
    }

    public void setContentStream(InputStream contentStream) {
        contentModified = true;

        // When a content is set as an input stream, prevoius contents set as object get invalidated
        // But we don't want to generate that content till it is explicitly get using getContent
        if (contentStream != null) {
            this.content = null;
        }

        this.contentStream = contentStream;
    }

    public Object getContent() throws RegistryException {
        if (content == null) {

            contentStream = getContentStream();
            if (contentStream != null) {
                BufferedInputStream bufferedInputStream = new BufferedInputStream(contentStream);
                ByteArrayOutputStream out = new ByteArrayOutputStream();
                try {
                    byte[] contentChunk = new byte[1024];
                    int byteCount;
                    while ((byteCount = bufferedInputStream.read(contentChunk)) != -1) {
                        out.write(contentChunk, 0, byteCount);
                    }
                    content = out.toByteArray();
                } catch (IOException e) {
                    String msg = "Failed to generate byte array from the resource content stream.";
                    log.error(msg, e);
                    throw new RegistryException(msg, e);
                } finally {
                    try {
                        bufferedInputStream.close();
                        out.close();
                    } catch (IOException e) {
                        log.error("Could not close the input stream of the content.", e);
                    }
                }
            }
        }

        return content;
    }

    public void setContent(Object content) {
        this.content = content;

        // When a new content is set, previous contents set as input stream get invalid
        // But we don't want to generate that content till it is explicitly get
        // using getContentStream
        if (content != null) {
            this.contentStream = null;
        }

        this.contentModified = true;
    }

    private String preparePath(String path) {

        String preparedPath = path;

        // make sure that the path does not end with a "/"
        if (!path.equals(RegistryConstants.ROOT_PATH) &&
                path.endsWith(RegistryConstants.PATH_SEPARATOR)) {
            preparedPath =
                    path.substring(0, path.length() - RegistryConstants.PATH_SEPARATOR.length());
        }

        return preparedPath;
    }


    public String getLastUpdaterUserName() {
        return lastUpdaterUserName;
    }

    public void setLastUpdaterUserName(String lastUpdaterUserName) {
        this.lastUpdaterUserName = lastUpdaterUserName;
    }

    public boolean isDirectory() {
        return directory;
    }

    public String[] getDependents() {
        return dependents.toArray(new String[dependents.size()]);
    }

    public void setDependents(String[] dependents) {
        this.dependents = new ArrayList();

        this.dependents.addAll(Arrays.asList(dependents));
    }

    public void addDependent(String dependentPath) {
        dependents.add(dependentPath);
    }

    public String[] getDependencies() {
        return dependencies.toArray(new String[dependencies.size()]);
    }

    public void setDependencies(String[] dependencies) {
        this.dependencies = new ArrayList();

        this.dependencies.addAll(Arrays.asList(dependencies));
    }

    public void addDependency(String dependencyPath) {
        dependencies.add(dependencyPath);
    }

    public boolean isContentModified() {
        return contentModified;
    }

    /**
     * This method is used to explicitly set the content modified state of the resource. Normal
     * users of the Registry API should not call this method.
     *
     * @param contentModified true if we want to add a new version upon putting this to the
     * registry. false otherwise.
     */
    public void setContentModified(boolean contentModified) {
        this.contentModified = contentModified;
    }
}
